
rm(list = ls()) # cleanup

# Load necessary packages

library(robustbase)
library(ellipse) 
library(svd) # for propack.svd
library(cellWise)
library(grid) # to plot images etc.
library(gridExtra) # has grid.arrange()
library(matrixStats)

##########################
## Examples of Wrapping ##
##########################

# This script contains 5 examples of the wrapping procedure.
# The first 2 examples, on the stars and bushfire datasets,
# illustrate robust correlation/covariance estimation 
# through wrapping, with anomaly (outlier) detection. 
# The third example shows robust PCA on a high-dimensional 
# video dataset.
# The next examples illustrate the application of wrapping 
# correlation to speed up the DetectDeviatingCells algorithm.
# At the end the figures about distance correlation are
# reproduced.


#################
# 1. Stars data #
#################

# load the data
#
?robustbase::starsCYG # background
X <- robustbase::starsCYG; plot(X,pch=16)

# Estimate location and covariance
#
cutoff <- sqrt(qchisq(0.99, ncol(X))); cutoff
X.mean <- colMeans(X)
X.cov  <- cov(X)
round(cor(X), 2) # -0.21

estX    <- estLocScale(X)
Xw      <- wrap(X, estX$loc, estX$scale)$Xw
Xw.mean <- colMeans(Xw)
Xw.cov  <- cov(Xw)
round(cor(Xw), 2) # 0.57

# Draw classical and wrapped tolerance ellipse
#
pdf("stars_ellipses.pdf",width=5.4,height=5.4) 
# sizes in inches
ell1 <- ellipse(X.cov,centre=X.mean,t=cutoff,npoints = 120)
ell2 <- ellipse(Xw.cov,centre=Xw.mean,t=cutoff,npoints=120)
xmin <- min(min(X[, 1]), min(ell1[, 1]), min(ell2[, 1])) - 0.25
xmax <- max(max(X[, 1]), max(ell1[, 1]), max(ell2[, 1])) + 0.5
ymin <- min(min(X[, 2]), min(ell1[, 2]), min(ell2[, 2])) - 0.5
ymax <- max(max(X[, 2]), max(ell1[, 2]), max(ell2[, 2])) + 0.25
myxlim <- c(xmin, xmax); myylim <- c(ymin, ymax)
plot(X, xlab = "",ylab = "",main = "",pch = 16, 
     col = "black", xlim = myxlim, ylim = myylim)
title(main="Stars: classical and wrapping ellipse",
            line=1) # , cex.lab=1.2, family="Calibri Light")
title(xlab="log(temperature)", line=2.3)
title(ylab="log(light intensity)", line=2.3)
points(ell1, type = "l", lwd = 1.5, col = "red")
points(ell2, type = "l", lwd = 1.5, col = "blue")
legend("bottomright",c("Classical","Wrapping"),
       col=c("red","blue"),lty=c(1,1),lwd=2)
dev.off()

# Compute and plot classical Mahalanobis distances
md <- sqrt(mahalanobis(X, center = X.mean, cov = X.cov))
#
pdf("stars_md.pdf",width=4,height=4.5) 
plot(1:nrow(X), md, pch = 16, cex = 0.9,
     ylab = "",xlab = "",main = "",xlim = c(0, nrow(X)), 
     ylim = c(0, max(md, cutoff) + 0.2), col = "black")
title(main="Mahalanobis distances",line=1)
title(xlab="case number", line=2.3)
title(ylab="Mahalanobis distance", line=2.3)
abline(h = cutoff, lwd = 1.5, col = "red")
# We only see some mild outliers.
dev.off()

# Compute and plot wrapping-based robust distances
#
rd <- sqrt(mahalanobis(X, center = Xw.mean, cov = Xw.cov))
#
pdf("stars_rd.pdf",width=4,height=4.5) 
plot(1:nrow(X), rd, pch = 16, cex = 0.9,
     ylab = "",xlab = "",main = "",xlim = c(0, nrow(X)), 
     ylim = c(0, max(rd, cutoff) + 0.2), col = "black")
title(main="Robust distances",line=1)
title(xlab="case number", line=2.3)
title(ylab="robust distance", line=2.3)
abline(h = cutoff, lwd = 1.5, col = "red")
# Here the four giant stars stand out.
dev.off()


####################
# 2. Bushfire data #
####################

# Load the data
#
?robustbase::bushfire
X <- as.matrix(robustbase::bushfire); pairs(X)

# Estimate location and covariance
#
cutoff <- sqrt(qchisq(0.975, ncol(X))); cutoff
X.mean <- colMeans(X)
X.cov  <- cov(X)
round(cor(X), 2)

estX    <- estLocScale(X)
Xw      <- wrap(X, estX$loc, estX$scale)$Xw
Xw.mean <- colMeans(Xw)
Xw.cov  <- cov(Xw)
round(cor(Xw), 2)

# Reweighting step
wd <- sqrt(mahalanobis(X, center = Xw.mean, cov = Xw.cov))
weights  <- (wd < cutoff) + 0
wmeancov <- cov.wt(X, weights, method = "ML")
X.wmean  <- wmeancov$center
X.wcov   <- wmeancov$cov
round(cov2cor(X.wcov), 2)

# Plot classical Mahalanobis distances
# 
md <- sqrt(mahalanobis(X, center = X.mean, cov = X.cov))
plot(1:nrow(X), md, pch = 16, cex = 0.9,
     ylab = "Mahalanobis distance", xlab = "Index",
     main = "Mahalanobis distances", xlim = c(0, nrow(X)),
     ylim = c(0, max(md, cutoff) + 0.2), col = "black")
abline(h = cutoff, lwd = 1.5, col = "red")
# No outliers stand out here, only borderline cases.

# Plot wrapping-based robust distances
# 
rwd <- sqrt(mahalanobis(X, center = X.wmean, cov = X.wcov))
plot(1:nrow(X),rwd,pch = 16, cex = 0.9,
     ylab = "robust distance", xlab = "Index",
     main = "Wrapping-based robust distances", 
     xlim = c(0, nrow(X)),
     ylim = c(0, max(rwd, cutoff) + 0.2), col = "black")
abline(h = cutoff, lwd = 1.5, col = "red")
# Many outliers visible, the same as in Maronna-Yohai 1995.

# Distance-distance plot:
#
plot(md, rwd, pch = 16, cex = 0.9, ylab = "robust distance",
     xlab = "classical distance", 
     main = "distance-distance plot",
     xlim = c(0, max(md, cutoff) + 0.2),
     ylim = c(0, max(rwd, cutoff) + 0.2), col = "black")
abline(v = cutoff, lwd = 1.5, col = "red")
abline(h = cutoff, lwd = 1.5, col = "red")
abline(0, 1, lwd = 1.5, lty = 2)
# Only the robust distances detect the outliers.

# Classical and wrapped tolerance ellipses:
#
coordgrid <- expand.grid(1:dim(X)[2], 1:dim(X)[2])
coordgrid <- coordgrid[-(1+(1:dim(X)[2]-1)*(dim(X)[2]+1)),]
i = 1
pairs(X, panel = function(x, y) {
  points(x, y)
  inds <- as.numeric(coordgrid[i, ])
  ell1 <- ellipse::ellipse(X.cov[inds, inds], 
                           centre = X.mean[inds],
                           t = cutoff, npoints = 120)
  ell2 <- ellipse::ellipse(X.wcov[inds, inds], 
                           centre = X.wmean[inds],
                           t = cutoff, npoints = 120)
  points(ell1, type = "l", lwd = 1.5, col = "red")
  points(ell2, type = "l", lwd = 1.5, col = "blue")
  i <<- i + 1
})
# The classical ellipses tend to be bigger.


########################
# 3. Parking lot video #
########################

# This is an example in the paper.

rm(list = ls())

# Download the data file parking_lot.RData from
# https://wis.kuleuven.be/stat/robust/papers-since-2010
# under the link [parking_lot.Rdata (big)]

# Load the data into R:
load("./parking_lot.Rdata")
# size of X in R: 1.6 Gb
dims <- dim(X); dims # 230 480 640   3
# This is a video with 230 frames, each with 480 by 640
# pixels, and each pixel in Red/Green/Blue.

# Look at a frame:
aframe <- X[90,,,];
aframe <- pmax(0,pmin(aframe,1)) 
dim(aframe) <- dims[2:4]; grid::grid.newpage()
grid::grid.raster(aframe,interpolate = FALSE)

# Combine all pixels of an image in a row:
#
dim(X) <- c(dims[1], prod(dims[2:4]))
dim(X) # 230 921600 # ultra-high dimension

# Estimate location and scale using estLocScale:
#
timestamp()
estX <- estLocScale(X, precScale = 1e-12)
timestamp() # 35 seconds
# Warning message:
#   In estLocScale(X, precScale = 1e-12) :
#  3137  out of  921600  variables have an estimated scale <= 
#   "precScale" =  1e-12 .

# Preprocess the data to deal with zero scales:
#
zeropix <- which(estX$scale < 1e-12)
nzero   <- length(zeropix); nzero 
filler  <- 2 * ((1:dims[1]) - (dims[1] + 1)/2)/(dims[1]-1)
filler  <- 0.9 / 255 * matrix(rep(filler, nzero), 
															ncol = nzero, byrow = F)
centers <- apply(X[, zeropix], 2, FUN = median)
centers <- pmax(0, pmin(1, centers)); 
filler  <- sweep(filler, 2, centers, "+")
X[,zeropix] <- filler
estX$scale[zeropix] <- estLocScale(X[,zeropix])$scale

rm(filler, aframe, centers, nzero, zeropix); gc()

# Wrap the data
#
timestamp()
Xw <- wrap(X, estX$loc, estX$scale)$Xw 
timestamp() # 6 seconds

# center the wrapped data
#
aveXw <- colMeans(Xw)
XwC   <- sweep(Xw, 2, aveXw)
rm(Xw) # to save memory

# Compute SVD of centered wrapped data
#
prcXwz    <- svd::propack.svd(XwC, neig = 3) # 12 seconds
rloadings <- prcXwz$v; dim(rloadings) # 921600  3
Xwscores  <- XwC %*% rloadings; dim(Xwscores) # 230 3
rm(XwC) # to save memory

## Project original data on robust loadings:
#
XC        <- sweep(X, 2, aveXw)
robscores <- XC %*% rloadings
rm(XC) # to save memory

# Compute the fitted frames and residuals:
#
X.rfitted <- robscores %*% t(rloadings[, 1:3])
X.rfitted <- sweep(X.rfitted, 2, aveXw, "+")
Xres      <- X - X.rfitted
rm(X.rfitted) # to save memory

# Plot the loading images for inspection
#
prepLoading <- function(v, dims) {
	dim(v) <- dims
	v <- apply(v, c(1, 2), sum)
	maxab <-  0.5 * max(abs(c(max(v), min(v))))
	v <- pmin(pmax(v, -maxab), maxab)
	return(list(loading = v, maxab = maxab))
}
Pal <- colorRampPalette(c("blue", "white","red"))(125)

PC1 <- prepLoading(rloadings[, 1], dims[2:4])
PC2 <- prepLoading(rloadings[, 2], dims[2:4])
PC3 <- prepLoading(rloadings[, 3], dims[2:4])

image((t(apply((-PC1$loading), 2, rev))), 
			zlim = PC1$maxab * c(-1, 1), col = Pal) # first loading
image((t(apply((PC2$loading), 2, rev))), 
			zlim = PC2$maxab * c(-1, 1), col = Pal)
image((t(apply((PC3$loading), 2, rev))), 
			zlim = PC3$maxab * c(-1, 1), col = Pal)
dev.off()

# These loadings model the background and not the people
# who are moving in front of it.

# Compute location and scale of the residuals
# and compute standardized squared residuals
locScale_res <- estLocScale(Xres)
#
# Plot the masked frames
#
showRes = function(i, Xres, locs = locScale_res$loc,
                      scales = locScale_res$scale, cutoff = 200) {
  restemp <- (Xres[i, ] - locs) / scales;
  dim(restemp) <- c(prod(dims[2:3]), dims[4]); # 307200 3
  res2 <- restemp[, 1]^2 + restemp[, 2]^2 + restemp[, 3]^2
  mask <- rep(0, prod(dims[2:3])); mask[res2 > cutoff] <- 1
  mask <- cbind(mask,mask,mask); dim(mask) <- dims[2:4]
  realim <- X[i, ];
  realim <- pmax(0,pmin(realim,1)); dim(realim) <- dims[2:4]
  masked <- realim * mask; dim(masked)
  masked <- pmax(0,pmin(masked,1)); dim(masked) <- dims[2:4]
  masked[masked == 0] <- 0.5
  finalim <- array(0, dim = c(dims[2], dims[3], dims[4]))
  finalim[, 1:dims[3], ] = masked
  # filename = paste("masked_",i,".pdf",sep = "")
  # pdf(filename,width = 64, height = 48)
  grid::grid.newpage()
  grid::grid.raster(finalim,interpolate = FALSE)
  # dev.off()
}
showRes(i=100,Xres)

rm(estX, locScale_res, PC1, PC2, PC3, Xres, rloadings, 
	 robscores, Xwscores, aveXw, prcXwz); gc()


# For comparison, we can repeat the above analysis using 
# classical PCA, PCA based on Spearman correlation, and
# PCA based on Huber's psi function. This takes a while.


## classical PCA
################

aveX <- colMeans(X)
Xz <- sweep(X, 2, aveX) 
prcXz <- svd::propack.svd(Xz, neig = 3) 
loadings <- prcXz$v
scores <- Xz %*% loadings
rm(Xz)
X.fitted <- scores %*% t(loadings[, 1:3])
X.fitted <- sweep(X.fitted, 2, aveX, "+")
Xres     <- X - X.fitted
rm(X.fitted)

# Plot the loading images for inspection:

PC1  <- prepLoading(loadings[, 1], dims[2:4])
PC2  <- prepLoading(loadings[, 2], dims[2:4])
PC3  <- prepLoading(loadings[, 3], dims[2:4])

image((t(apply((PC1$loading), 2, rev))), 
			zlim = PC1$maxab * c(-1, 1), col = Pal)
image((t(apply((PC2$loading), 2, rev))), 
			zlim = PC2$maxab * c(-1, 1), col = Pal)
image((t(apply((PC3$loading), 2, rev))), 
			zlim = PC3$maxab * c(-1, 1), col = Pal)

# These loadings are not robust: they are 
# strongly attracted by the men in the frames.

# Plot the masked frames
#
locs <- colMeans(Xres)
scales <- colSds(Xres)
showRes(100, Xres, locs, scales, 20)
# This masked frame is more noisy than with wrapping.

rm(PC1, PC2, PC3,loadings, scores, Xres, aveX, prcXz,
	 locs, scales); gc()


## PCA by Spearman correlation
##############################

# We again use the matrix X from above, where columns
# with zero scales have been filled.

# Compute ranks by column:
# Xr <- scale(apply(X, 2, rank)) # slow
# More efficiently:
Xr <- matrixStats::colRanks(X,ties.method="average",preserveShape=T)
aveXr <- matrixStats::colMeans2(Xr)
summary(aveXr) # only identical if ties.method="average"
Xr <- sweep(Xr, 2, aveXr); dim(Xr)  # centered ranks
summary(matrixStats::colMeans2(Xr)) # zeroes

# Rescale Xr to original sd:
scalefac = matrixStats::colSds(X)/matrixStats::colSds(Xr)
Xr <- sweep(Xr, 2, scalefac, "*"); dim(Xr)
summary(matrixStats::colMeans2(Xr)) # zeroes
summary(matrixStats::colSds(Xr))    # is same as:
summary(matrixStats::colSds(X))     # original scales.

prcXr <- svd::propack.svd(Xr, neig = 3) 
rm(Xr)
loadings <- prcXr$v
# medX <- apply(X, 2, median) # slow, better:
medX <- matrixStats::colMedians(X)
Xz <- sweep(X, 2, medX)
scores <- Xz %*% loadings
rm(Xz)
X.fitted <- scores %*% t(loadings[, 1:3])
X.fitted <- sweep(X.fitted, 2, medX, "+")
Xres     <- X - X.fitted
rm(prcXr, X.fitted)

# Plot the loading images for inspection
#
PC1  <- prepLoading(loadings[, 1], dims[2:4])
PC2  <- prepLoading(loadings[, 2], dims[2:4])
PC3  <- prepLoading(loadings[, 3], dims[2:4])

image((t(apply((PC1$loading), 2, rev))), 
			zlim = PC1$maxab * c(-1, 1), col = Pal)
image((t(apply((PC2$loading), 2, rev))), 
			zlim = PC2$maxab * c(-1, 1), col = Pal)
image((t(apply((PC3$loading), 2, rev))), 
			zlim = PC3$maxab * c(-1, 1), col = Pal)

# These loadings are less robust than wrapping

# Plot the masked frames
#
locs   <- matrixStats::colMedians(Xres)
scales <- matrixStats::colMads(Xres)
showRes(100, Xres, locs, scales, 50)

rm(PC1, PC2, PC3, loadings, scores, Xres, medX, locs, scales); gc()


## PCA with Huber psi function
##############################

# Estimate location and scale with Huber's psi and rho:
timestamp()
locScale <- cellWise::estLocScale(X,type="hubhub",precScale=1e-12)
timestamp() # 12 seconds
locs     <- locScale$loc
scales   <- locScale$scale

# Apply Huber's psi to the data:

psiH = function(y,b){
  pmin(pmax(y,-b),b)
}

Xw = scale(X, locs, scales); dim(Xw)
Xw = apply(Xw, 2, function(y) psiH(y,b=1.5))
# We need centered transformed data, so we only put
# the scales back:
Xw = scale(Xw,center=F,scale=1/scales)

# Compute SVD of centered transformed data
prcXwz    <- svd::propack.svd(Xw, neig = 3) # 12 seconds
rloadings <- prcXwz$v; dim(rloadings) # 921600  3
Xwscores  <- Xw %*% rloadings; dim(Xwscores) # 230 3
rm(Xw) # to save memory

## Project original data on robust loadings:
#
XC        <- sweep(X, 2, locs)
robscores <- XC %*% rloadings
rm(XC) # to save memory

# Compute the fitted frames and residuals:
#
X.rfitted <- robscores %*% t(rloadings[, 1:3])
X.rfitted <- sweep(X.rfitted, 2, locs, "+")
Xres      <- X - X.rfitted
rm(X.rfitted) # to save memory

# Plot the loading images for inspection
PC1 <- prepLoading(rloadings[, 1], dims[2:4])
PC2 <- prepLoading(rloadings[, 2], dims[2:4])
PC3 <- prepLoading(rloadings[, 3], dims[2:4])

# filename = paste("PC1_Huber.pdf",sep = "")
# pdf(filename,width = 64, height = 48)
image((t(apply((-PC1$loading), 2, rev))), 
      zlim = PC1$maxab * c(-1, 1), col = Pal)
# dev.off()

# Compute location and scale of the residuals
# and compute standardized squared residuals
locScale_res <- estLocScale(Xres, type = "hubhub")
#
# filename = paste("residuals_Huber.pdf",sep = "")
# pdf(filename,width = 64, height = 48)
showRes(100, Xres)
# dev.off()

rm(locScale, locScale_res, PC1, PC2, PC3, Xres, rloadings, 
   robscores, Xwscores, locs, scales, prcXwz, filename)


#################
# 4. Glass data #
#################

rm(list = ls())

# Load the data:
#
data(glass)
dim(glass) # 180 750
sum(is.na(glass)) # 0 # no NA's

# DetectDeviatingCells (DDC)

# Compute the regular and fast DDC 
#
DDCpars=list(fastDDC=F,silent=F)
timestamp()
DDCglass =  DDC(glass,DDCpars) 
timestamp() # for silent=T : 9 seconds
names(DDCglass)
remX = DDCglass$remX; dim(remX) # 180 737

fastDDCpars=list(fastDDC=T,silent=F)
timestamp()
fastDDCglass = DDC(glass,fastDDCpars)
timestamp() # for silent=T : 1 second

# plot settings
n = nrow(remX)
nrowsinblock = 5
rowlabels = rep("",floor(n/nrowsinblock));
rowlabels[1] = "1"
rowlabels[floor(n/nrowsinblock)] = "n";
rowlabels
rowtitle = "glass samples"

d = ncol(remX)
ncolumnsinblock = 5
columnlabels = rep("",floor(d/ncolumnsinblock));
columnlabels[1] = "1";
columnlabels[floor(d/ncolumnsinblock)] = "d"
columnlabels
columntitle = "wavelengths"

# plot cell maps:
ggpDDC = cellMap(D=remX, 
								 R=DDCglass$stdResid,
								 indcells=DDCglass$indcells,
								 indrows=DDCglass$indrows,
								 rowlabels=rowlabels,
								 columnlabels=columnlabels,
								 mTitle="fastDDC=F",
								 nrowsinblock=nrowsinblock,
								 ncolumnsinblock=ncolumnsinblock,
								 columnangle=0,
								 rowtitle=rowtitle,
								 columntitle=columntitle,
								 sizetitles=1.5,
								 autolabel=F)
plot(ggpDDC)


ggpfastDDC = cellMap(D=remX, 
										 R=fastDDCglass$stdResid,
										 indcells=fastDDCglass$indcells,
										 indrows=fastDDCglass$indrows,
										 rowlabels=rowlabels,
										 columnlabels=columnlabels,										 
										 mTitle="fastDDC=T",
										 nrowsinblock=nrowsinblock,
										 ncolumnsinblock=ncolumnsinblock,
										 columnangle=0,
										 rowtitle=rowtitle,
										 columntitle=columntitle,
										 sizetitles=1.5,
										 autolabel=F)
plot(ggpfastDDC) 

# Creating the pdf:
# pdfName = "cellmaps_glass.pdf"
# pdf(pdfName, width = 16, height = 10 )
grid.arrange(ggpDDC,ggpfastDDC,ncol=1) 
# dev.off()

# The results are very similar, but not identical.


####################
# 5. Prostate data #
####################

# Download the file Singh.rda from
# http://www.stats.uwo.ca/faculty/aim/2015/9850/
# microarrays/FitMArray/data/
# rm(list = ls())

# load the data
load("./Singh.rda")
summary(Singh)
X  = Singh$X; dim(X)     # 102 12600
Xt = Singh$Xt; dim(Xt)   #  34 12600
X  = rbind(X,Xt); dim(X) # 136 12600
rownames(X) = NULL # to avoid duplicate names
dim(X) # 136 12600

scales = apply(X,2,mad); length(scales) # 12600
summary(scales)
#  Min.  1st Qu.   Median     Mean  3rd Qu.     Max. 
# 1.483    7.413   14.826   32.302   31.135 3297.302 
# so the minimum is strictly positive
# ==> DDC will not remove columns

## Do NOT try to run with fastDDC=F, would take very long!
# With fastDDC=T :
fastDDCpars = list(fastDDC=TRUE)
timestamp()
DDCout = DDC(X,fastDDCpars) 
# The input data has 136 rows and 12600 columns.
timestamp() # less than a minute

remX = DDCout$remX; dim(remX) # 136 12600
stdres = DDCout$stdResid; dim(stdres) # same
indrows = DDCout$indrows; length(indrows) # 33 rows flagged
indrows 
#  [1] 103 104 105 106 108 109 110 111 112 113 114 115 116 117
# [15] 118 119 120 121 122 123 124 125 126 127 128 129 130 131
# [29] 132 133 134 135 136
# This is the test data minus one patient! (107)

# plot row outlyingness Ti
plot(DDCout$Ti)
cutoff = sqrt(qchisq(0.99,1)); cutoff # 2.575829
abline(h=cutoff)
abline(v=102.5) # The test data is to the right.

# Let's count the number of flagged cells per column:
countflagged = function(z){ sum(abs(z) > cutoff) }
numflagged = as.vector(apply(stdres,2,FUN=countflagged)) 
length(numflagged) # 12600
hist(numflagged)
indcols = which(numflagged > 49); length(indcols) # 92
# Could not get exactly 100 due to ties.
hist(indcols) # roughly uniformly all over the place...

# Make a cellmap of the most outlying variables (genes):
tempXrem = remX[,indcols]; dim(tempXrem) 
tempXres = stdres[,indcols]; dim(tempXres)
tempindcells = which(abs(tempXres) > cutoff)
length(tempindcells) 
nrowsinblock = 1
ncolumnsinblock = 1
myrowlabels = rep("",nrow(tempXres)); length(myrowlabels)
mycolumnlabels = rep("",ncol(tempXres)); length(mycolumnlabels)

ggpDDC = cellMap(D=tempXrem, 
								 R=tempXres,
								 indcells=tempindcells,
								 indrows=NULL,
								 rowlabels=myrowlabels,
								 columnlabels=mycolumnlabels,
								 mTitle="Genes in prostate data")

# pdf(paste("./cellmap_prostate.pdf",sep=""),width=7,height=6)
plot(ggpDDC)
# dev.off()

################################################################

## Figures about distance correlation
#####################################

rm(list = ls())
library(energy)      # for dcor, dcor.test, dcor2d
library(matrixStats) # for colMedians, colMads


## Figure 8, left panel: Empirical power for t(1) data
## with d=2 and X_1 = X_2, as a function of sample size 

m = 50 # number of replications was 1000 in paper, takes longer.
samplesizes = (1:20)*10
maxn = max(samplesizes) # 200
result_dcor = array(0,dim = c(m,length(samplesizes),2))
result_dcorTest = array(0,dim = c(m,length(samplesizes),2))
for (i in 1:m) {
  set.seed(i)
  Xuni = rt(maxn,df=1)
  X1 = cbind(Xuni,Xuni) # linearly dependent
  X2 = matrix(rt(maxn*2,df=1),maxn,2) # independent
  X = cbind(X1[,1],X2[,1])
  Y = cbind(X1[,2],X2[,2])
  for (j in 1:length(samplesizes)) { 
    # j = 1
    n = samplesizes[j]
    Xn = X[1:n,]
    Yn = Y[1:n,]
    result_dcor[i,j,1] = dcor(Xn,Yn)
    result_dcorTest[i,j,1] = dcor.test(Xn,Yn,R=1000)$p.value
    # For robust version must first standardize the variables:
    Xn[,1] = scale(Xn[,1],center=median(Xn[,1]),scale=mad(Xn[,1]))
    Xn[,2] = scale(Xn[,2],center=median(Xn[,2]),scale=mad(Xn[,2]))
    Yn[,1] = scale(Yn[,1],center=median(Yn[,1]),scale=mad(Yn[,1]))
    Yn[,2] = scale(Yn[,2],center=median(Yn[,2]),scale=mad(Yn[,2]))    
    tX = cbind(tanh(Xn[,1]),tanh(Xn[,2]))
    tY = cbind(tanh(Yn[,1]),tanh(Yn[,2]))
    result_dcor[i,j,2] = dcor(tX,tY)
    result_dcorTest[i,j,2] = dcor.test(tX,tY,R=1000)$p.value
  }
  print(i)
}
powers = apply(result_dcorTest<0.1,2,function(y) colSums(y))/m
dcorpower = powers[1,]
robcorpower = powers[2,]

pdf("Figure8_left.pdf",width=5.4,height=5.4)
xmin = 10
xmax = 200
ymin = 0
ymax = 1
myxlim = c(xmin,xmax); myylim = c(ymin,ymax)
placeholder = rbind(c(xmin,ymin),c(xmax,ymax))
plot(placeholder,xlab="",ylab="",main="",type="n",pch=16,
     col="black",xlim=myxlim,ylim=myylim)
title(main="Empirical power at 0.1 significance",line=1) 
title(xlab="sample size n",line=2.3)
title(ylab="power",line=2.3)
points(samplesizes,dcorpower,type ="l",lwd=1.5,col="black")
points(samplesizes,robcorpower,type ="l",lwd=1.5,col="blue")
dev.off()


## Figure 8, right panel: Empirical power for t(1) data
## with n=100 and X_1 = X_2, as a function of dimension d

m = 50 # was 1000 in paper, takes longer.
n = 100 # fixed sample size
maxd = 30 # maximal number of dimensions
result_dcor = array(0,dim = c(m,maxd,2))
result_dcorTest = array(0,dim = c(m,maxd,2))
for (i in 1:m) {
  set.seed(i)
  Xuni = rt(n,df=1)
  Xextra = matrix(rt(n*(maxd-1),df=1),n,maxd-1) # independent
  Yextra = matrix(rt(n*(maxd-1),df=1),n,maxd-1) # independent 
  X = cbind(Xuni,Xextra)
  Y = cbind(Xuni,Yextra)
  XcolMed = colMedians(X)
  XcolMad = colMads(X)  
  YcolMed = colMedians(Y)
  YcolMad = colMads(Y)    
  for (j in 1:maxd) { 
    Xd = X[,1:j]
    Yd = Y[,1:j]
    result_dcor[i,j,1] = dcor(Xd,Yd)
    result_dcorTest[i,j,1] = dcor.test(Xd,Yd,R=1000)$p.value
    # For robust version must first standardize the variables:
    Xd = scale(Xd,center=XcolMed[1:j],scale=XcolMad[1:j])
    Yd = scale(Yd,center=YcolMed[1:j],scale=YcolMad[1:j])    
    result_dcor[i,j,2] = dcor(tanh(Xd),tanh(Yd))
    result_dcorTest[i,j,2] = dcor.test(tanh(Xd),tanh(Yd),R=1000)$p.value
  }
  print(i)
}
powers = apply(result_dcorTest<0.1,2,function(y) colSums(y))/m
dcorpower = powers[1,]
robcorpower = powers[2,]

pdf("Figure8_right.pdf",width=5.4,height=5.4)
xmin = 1
xmax = 30
ymin = 0
ymax = 1
myxlim = c(xmin,xmax); myylim = c(ymin,ymax)
placeholder = rbind(c(xmin,ymin),c(xmax,ymax))
plot(placeholder,xlab="",ylab="",main="",type="n",pch=16,
     col="black",xlim=myxlim,ylim=myylim)
title(main="Empirical power at 0.1 significance",line=1) 
title(xlab="dimension d",line=2.3)
title(ylab="power",line=2.3)
points(1:30,dcorpower,type ="l",lwd=1.5,col="black")
points(1:30,robcorpower,type ="l",lwd=1.5,col="blue")
dev.off()


# Figure 16, left panel: make dcor approach 1 by a far outlier
##############################################################

# generate bivariate standard normal data:
n = 100000
set.seed(123)
X = MASS::mvrnorm(n,rep(0,2),diag(2)); dim(X) # 100000  2
m = 50 # for plot resolution, was 200 in paper (takes longer)
avalues = ((0:m)/m)*2e5
dcors  = rep(NA,m+1)
robcors = rep(NA,m+1)
Xc = X
for (i in 1:(m+1)){
  Xc[1,]     = rep(avalues[i],2) # replaces 1 data point by (a,a)
  dcors[i]  = dcor2d(Xc[,1],Xc[,2])
  Xc[,1]     = scale(Xc[,1],center=median(Xc[,1]),scale=mad(Xc[,1]))
  Xc[,2]     = scale(Xc[,2],center=median(Xc[,2]),scale=mad(Xc[,2]))
  robcors[i] = dcor2d(tanh(Xc[,1]),tanh(Xc[,2]))
  print(i)
}

pdf("Figure16_left.pdf",width=5.4,height=5.4)
xmin = 0
xmax = avalues[m+1]
ymin = 0
ymax = 1
myxlim = c(xmin,xmax); myylim = c(ymin,ymax)
placeholder = rbind(c(xmin,ymin),c(xmax,ymax))
# plot(placeholder,xlab="",ylab="",main="",type="n",pch=16,
#      col="black",xlim=myxlim,ylim=myylim)
# To change the tick mark labels on x-axis:
# suppress the x-axis:
plot(placeholder,xlab="",ylab="",main="",type="n",pch=16,
     col="black",xlim=myxlim,ylim=myylim,xaxt="n")
# placing x-axis tick labels:
xtick=seq(0,200000,by=50000)
axis(side=1,at=xtick,labels=FALSE)
axis(1,at=xtick,labels=c("0e+00","5e+04","1e+05","1.5e+05","2e+05"))
title(main="dCor of binormal data with 1 outlier at (a,a)",line=1) 
title(xlab="a",line=2.3)
title(ylab="dCor",line=2.3)
points(avalues,dcors,type ="l",lwd=1.5,col="black")
points(avalues,robcors,type ="l",lwd=1.5,col="blue")
dev.off()


# Figure 16, right panel: make dcor approach 0 by a far outlier
###############################################################

# Now start with Y=X:
n = 100000
set.seed(123)
X = rnorm(n); X = cbind(X,X)
m = 50 # for plot resolution, was 200 in paper
avalues = ((0:m)/m)*1e6
dcors   = rep(NA,m+1)
robcors = rep(NA,m+1)
Xc = X
for (i in 1:(m+1)){
  Xc[1,]     = c(avalues[i],0)
  dcors[i]   = dcor2d(Xc[,1],Xc[,2])
  Xc[,1]     = scale(Xc[,1],center=median(Xc[,1]),scale=mad(Xc[,1]))
  Xc[,2]     = scale(Xc[,2],center=median(Xc[,2]),scale=mad(Xc[,2]))
  robcors[i] = dcor2d(tanh(Xc[,1]),tanh(Xc[,2]))
  print(i)
}

pdf("Figure16_right.pdf",width=5.4,height=5.4) # sizes in inches
xmin = 0
xmax = avalues[m+1]
ymin = 0
ymax = 1
myxlim = c(xmin,xmax); myylim = c(ymin,ymax)
placeholder = rbind(c(xmin,ymin),c(xmax,ymax))
plot(placeholder,xlab="",ylab="",main="",type="n",pch=16,
     col="black",xlim=myxlim,ylim=myylim)
# Now the tickmark labels come out fine from the start.
title(main="dCor of linear data with 1 outlier at (a,0)",line=1) 
title(xlab="a",line=2.3)
title(ylab="dCor",line=2.3)
points(avalues,dcors,type ="l",lwd=1.5,col="black")
points(avalues,robcors,type ="l",lwd=1.5,col="blue")
dev.off()

#################################################################

